import enum
from datetime import datetime
from typing import List, Optional, Type

from pydantic import BaseModel


class AssetStatus(enum.StrEnum):
    rejected = enum.auto()
    upload_in_progress = enum.auto()
    transform_in_progress = enum.auto()
    enrich_in_progress = enum.auto()
    accepted = enum.auto()


class TranscodeMode(enum.StrEnum):
    original = enum.auto()
    image_original_size = enum.auto()
    image_fit_1024 = enum.auto()
    image_fit_640 = enum.auto()
    image_fit_224 = enum.auto()
    image_cover_1024 = enum.auto()
    image_cover_640 = enum.auto()
    image_cover_224 = enum.auto()


AutoAnnotate = str


class ModelType(enum.StrEnum):
    epdet_b1 = enum.auto()


class ModelStatus(enum.StrEnum):
    error = enum.auto()
    draft = enum.auto()
    requested = enum.auto()
    in_progress = enum.auto()
    available = enum.auto()
    published = enum.auto()


class AnnotationType(enum.StrEnum):
    ground_truth = enum.auto()
    prediction = enum.auto()
    auto = enum.auto()


class UserReview(enum.StrEnum):
    approved = enum.auto()
    rejected = enum.auto()
    unknown = enum.auto()


class DatasetVersionResponse(BaseModel):
    version: int
    created_at: datetime
    updated_at: datetime
    asset_count: int
    modifiable: bool
    hero_asset_uuid: Optional[str] = None


class DatasetResponse(BaseModel):
    uuid: str
    name: str
    description: Optional[str] = ""
    created_at: datetime
    updated_at: datetime
    tags: List[str]
    account_uuid: str
    auto_annotates: List[AutoAnnotate]
    versions: List[DatasetVersionResponse]


class DatasetCreate(BaseModel):
    name: str
    description: str = ""
    tags: List[str] = []
    auto_annotates: List[AutoAnnotate] = []


class DatasetUpdate(DatasetCreate):
    pass


class Point2d(BaseModel):
    x: float
    y: float


class Point3d(Point2d):
    z: Optional[float] = None


class Box(BaseModel):
    topLeft: Point2d
    bottomRight: Point2d


class PredictedClass(BaseModel):
    id: Optional[int] = None
    confidence: Optional[float] = None
    classLabel: str
    category: Optional[str] = None


class PredictedLabel(BaseModel):
    id: Optional[int] = None
    confidence: Optional[float] = None
    label: str
    category: Optional[str] = None


class Contour(BaseModel):
    points: List[Point2d]
    cutouts: List[List[Point2d]]


class Mask(BaseModel):
    bitmap: str
    width: int
    height: int
    stride: int


class PredictedMesh(BaseModel):
    id: Optional[int] = None
    category: Optional[str] = None
    confidence: Optional[float] = None
    points: List[Point3d]


class PredictedKeyPoint(Point3d, PredictedClass):
    visible: Optional[bool] = None


class PredictedKeyPoints(BaseModel):
    category: Optional[str] = None
    type: Optional[str] = None
    points: List[PredictedKeyPoint]


class PredictedObject(PredictedClass):
    traceId: Optional[int] = None
    x: float
    y: float
    width: float
    height: float
    orientation: Optional[float] = None
    outline: Optional[List[Point2d]] = None
    contours: Optional[List[Contour]] = None
    mask: Optional[Mask] = None
    objects: Optional[List["PredictedObject"]] = None
    classes: Optional[List[PredictedClass]] = None
    labels: Optional[List[PredictedLabel]] = None
    meshs: Optional[List[PredictedMesh]] = None
    keyPoints: Optional[List[PredictedKeyPoints]] = None


class Prediction(BaseModel):
    source_width: float
    source_height: float
    objects: Optional[List[PredictedObject]] = None
    classes: Optional[List[PredictedClass]] = None
    labels: Optional[List[PredictedLabel]] = None
    meshs: Optional[List[PredictedMesh]] = None
    keyPoints: Optional[List[PredictedKeyPoints]] = None


class AssetAnnotationResponse(BaseModel):
    type: AnnotationType
    user_review: UserReview
    auto_annotate: Optional[AutoAnnotate] = []
    annotation: Optional[Prediction] = None
    uncertainty_score: Optional[float] = None


class AssetResponse(BaseModel):
    uuid: str
    created_at: datetime
    updated_at: datetime
    mime_type: str
    file_size_bytes: Optional[int] = None
    status: Optional[AssetStatus] = None
    status_message: Optional[str] = None
    external_id: Optional[str] = None
    annotations: Optional[List[AssetAnnotationResponse]] = []


class AssetImport(BaseModel):
    url: str
    ground_truth: Optional[Prediction] = None


class ModelResponse(BaseModel):
    uuid: str
    created_at: datetime
    updated_at: datetime
    account_uuid: str
    dataset_uuid: str
    dataset_version: int
    name: str
    description: Optional[str] = None
    type: ModelType
    status: ModelStatus
    status_message: Optional[str] = None
    exported_url: Optional[str] = None


class ModelCreate(BaseModel):
    name: str
    description: Optional[str] = None
    type: ModelType


class ModelUpdate(ModelCreate):
    pass


import asyncio
import logging
import time
from pathlib import Path

from eyepop import EyePopSdk
from eyepop import Job

source_path = Path(__file__).resolve()
source_dir = source_path.parent
example_url_1 = 'rtsp://wowzaec2demo.streamlock.net/vod/mp4:BigBuckBunny_115k.mp4'

logging.basicConfig(level=logging.INFO)
logging.getLogger('eyepop').setLevel(level=logging.DEBUG)


async def async_load_from_rtmp(url: str):
    async def on_ready(job: Job):
        print('async_load_from_rtmp on_ready')
        while result := await job.predict():
            print(result)

    async with EyePopSdk.workerEndpoint(is_async=True) as endpoint:
        await endpoint.load_from(url, on_ready)


t1 = time.time()
asyncio.run(async_load_from_rtmp(example_url_1))
t2 = time.time()
print("1x video async: ", t2 - t1)



import json
import logging
import time
from io import StringIO
from typing import Callable, BinaryIO, Any
from urllib.parse import urljoin
import warnings

import aiohttp

from eyepop.endpoint import Endpoint
from eyepop.exceptions import PopNotStartedException, PopConfigurationException, PopNotReachableException
from eyepop.worker.worker_jobs import WorkerJob, _UploadJob, _LoadFromJob, _UploadStreamJob
from eyepop.worker.load_balancer import EndpointLoadBalancer
from eyepop.worker.worker_syncify import SyncWorkerJob

log = logging.getLogger('eyepop')
log_requests = logging.getLogger('eyepop.requests')
log_metrics = logging.getLogger('eyepop.metrics')

MIN_CONFIG_RECONNECT_SECS = 10.0
MAX_RETRY_TIME_SECS = 30.0
FORCE_REFRESH_CONFIG_SECS = (61.0 * 61.0)


class WorkerEndpoint(Endpoint):
    """
    Endpoint to an EyePop.ai worker.
    """

    def __init__(self, secret_key: str, eyepop_url: str, pop_id: str, auto_start: bool, stop_jobs: bool,
                 job_queue_length: int, is_sandbox: bool, request_tracer_max_buffer: int):
        super().__init__(secret_key, eyepop_url, job_queue_length, request_tracer_max_buffer)
        self.pop_id = pop_id
        self.auto_start = auto_start
        self.stop_jobs = stop_jobs
        self.is_sandbox = is_sandbox

        self.sandbox_id = None

        self.is_dev_mode = True

        self.worker_config = None
        self.last_fetch_config_success_time = None
        self.last_fetch_config_error = None
        self.last_fetch_config_error_time = None

        self.pop_comp = None
        self.post_transform = None
        self.add_retry_handler(404, self._retry_404)

    async def _retry_404(self, status_code: int, failed_attempts: int) -> bool:
        if failed_attempts > 1:
            return False
        else:
            log_requests.debug('after 404, about to retry with fresh config')
            self.worker_config = None
            return True

    async def _disconnect(self):
        if self.sandbox_id is not None:
            try:
                base_url = await self.dev_mode_base_url()
                delete_sandbox_url = f'{base_url}/sandboxes/{self.sandbox_id}'
                headers = {'Authorization': await self._authorization_header()}
                log_requests.debug('before DELETE %s', delete_sandbox_url)
                await self.client_session.delete(delete_sandbox_url, headers=headers)
                log_requests.debug('after DELETE %s', delete_sandbox_url)
            except Exception as e:
                log.exception("error at disconnect", e)
            finally:
                self.sandbox_id = None

    async def _reconnect(self):
        if self.worker_config is not None:
            return

        if self.last_fetch_config_error_time is not None and self.last_fetch_config_error_time > time.time() - MIN_CONFIG_RECONNECT_SECS:
            raise self.last_fetch_config_error

        if self.last_fetch_config_success_time is not None and self.last_fetch_config_success_time > time.time() - MIN_CONFIG_RECONNECT_SECS:
            raise aiohttp.ClientConnectionError()

        if self.pop_id == 'transient':
            config_url = f'{self.eyepop_url}/workers/config'
        else:
            config_url = f'{self.eyepop_url}/pops/{self.pop_id}/config?auto_start={self.auto_start}'
        headers = {'Authorization': await self._authorization_header()}
        try:
            log_requests.debug('before GET %s', config_url)
            async with self.client_session.get(config_url, headers=headers) as response:
                self.worker_config = await response.json()
                self.last_fetch_config_success_time = time.time()
                self.last_fetch_config_error = None
                self.last_fetch_config_error_time = None
        except aiohttp.ClientResponseError as e:
            if e.status != 401:
                self.last_fetch_config_error = e
                self.last_fetch_config_error_time = time.time()
                raise e
            else:
                log_requests.debug('after GET %s: 401, about to retry with fresh access token', config_url)
                self.token = None
                self.expire_token_time = None
                headers = {'Authorization': await self._authorization_header()}
                async with self.client_session.get(config_url, headers=headers) as retried_response:
                    self.worker_config = await retried_response.json()
        except aiohttp.ClientConnectionError as e:
            self.last_fetch_config_error = e
            self.last_fetch_config_error_time = time.time()
            raise e

        self.is_dev_mode = self.is_sandbox or self.pop_id == 'transient' or self.worker_config[
            'status'] != 'active_prod'

        log_requests.debug(f'after GET {config_url}: {self.worker_config}')

        base_url = await self.dev_mode_base_url()

        if self.is_sandbox and self.sandbox_id is None:
            create_sandbox_url = f'{base_url}/sandboxes'
            headers = {'Authorization': await self._authorization_header()}
            log_requests.debug('before POST %s', create_sandbox_url)
            async with self.client_session.post(create_sandbox_url, headers=headers) as response:
                response_json = await response.json()
            log_requests.debug('after POST %s', create_sandbox_url)
            self.sandbox_id = response_json

        if self.pop_id == 'transient':
            self.pop_comp = 'identity'
            if self.sandbox_id is None:
                start_pipeline_url = f'{base_url}/pipelines'
            else:
                start_pipeline_url = f'{base_url}/pipelines?sandboxId={self.sandbox_id}'

            body = {'inferPipelineDef': {'pipeline': self.pop_comp},
                    'postTransformDef': {'transform': self.post_transform}, "source": {"sourceType": "NONE", },
                    "idleTimeoutSeconds": 60, "logging": ["out_meta"], "videoOutput": "no_output"}

            headers = {'Authorization': await self._authorization_header()}
            log_requests.debug('before POST %s', start_pipeline_url)
            async with self.client_session.post(start_pipeline_url, headers=headers, json=body) as response:
                response_json = await response.json()
            log_requests.debug('after POST %s', start_pipeline_url)
            self.worker_config['pipeline_id'] = response_json['id']

        if self.is_dev_mode and (self.worker_config['base_url'] is None or self.worker_config['pipeline_id'] is None):
            raise PopNotStartedException(pop_id=self.pop_id)

        if self.is_dev_mode and self.stop_jobs:
            stop_jobs_url = f'{await self.dev_mode_pipeline_base_url()}/source?mode=preempt&processing=sync'
            body = {'sourceType': 'NONE'}
            headers = {'Authorization': await self._authorization_header()}
            log_requests.debug('before PATCH %s', stop_jobs_url)
            async with self.client_session.patch(stop_jobs_url, headers=headers, json=body) as response:
                pass
            log_requests.debug('after PATCH %s', stop_jobs_url)

        if self.is_dev_mode and self.pop_comp is None:
            # get current pipeline string and store
            get_url = await self.dev_mode_pipeline_base_url()
            headers = {'Authorization': await self._authorization_header()}
            log_requests.debug('before GET %s', get_url)
            async with self.client_session.get(get_url, headers=headers) as response:
                response_json = await response.json()
                self.pop_comp = response_json.get('inferPipeline')
                self.post_transform = response_json.get('postTransform')
            log_requests.debug('after GET %s', get_url)
            log.debug('current popComp is %s', self.pop_comp)

        if self.is_dev_mode:
            endpoint = {'base_url': urljoin(self.eyepop_url, self.worker_config['base_url']).rstrip("/"),
                'pipeline_id': self.worker_config['pipeline_id']}
            self.load_balancer = EndpointLoadBalancer([endpoint])
        else:
            self.load_balancer = EndpointLoadBalancer(self.worker_config['endpoints'])

    async def get_pop_comp(self) -> str:
        return self.pop_comp

    async def set_pop_comp(self, pop_comp: str = None):
        if not self.is_dev_mode:
            raise PopConfigurationException(self.pop_id, 'set_pop_comp not supported in production mode')
        response = await self.pipeline_patch('inferencePipeline', content_type='application/json',
                                             data=json.dumps({'pipeline': pop_comp}))
        self.pop_comp = pop_comp
        return response

    async def get_post_transform(self) -> str:
        return self.post_transform

    async def set_post_transform(self, transform: str = None):
        if not self.is_dev_mode:
            raise PopConfigurationException(self.pop_id, 'set_post_transform not supported in production mode')
        response = await self.pipeline_patch('postTransform', content_type='application/json',
                                             data=json.dumps({'transform': transform}))
        self.post_transform = transform
        return response

    '''
    Deprecated
    '''

    async def list_models(self) -> list[dict]:
        warnings.warn("list_models for development use only", DeprecationWarning)
        if self.sandbox_id is None:
            get_path = 'models/instances'
        else:
            get_path = f'models/instances?sandboxId={self.sandbox_id}'

        response = await self._worker_request_with_retry('GET', url_path_and_query=get_path)
        return await response.json()

    async def get_manifest(self) -> list[dict]:
        warnings.warn("get_manifest for development use only", DeprecationWarning)
        if self.sandbox_id is None:
            get_path = 'models/sources'
        else:
            get_path = f'models/sources?sandboxId={self.sandbox_id}'
        response = await self._worker_request_with_retry('GET', url_path_and_query=get_path)
        return await response.json()

    async def set_manifest(self, manifests: list[dict]):
        if not self.is_dev_mode:
            raise PopConfigurationException(self.pop_id, 'set_manifest not supported in production mode')
        warnings.warn("set_manifest for development use only", DeprecationWarning)
        if self.sandbox_id is None:
            put_path = 'models/sources'
        else:
            put_path = f'models/sources?sandboxId={self.sandbox_id}'
        await self._worker_request_with_retry('PUT', url_path_and_query=put_path, content_type='application/json',
                                              data=json.dumps(manifests))

    async def load_model(self, model: dict, override: bool = False) -> dict:
        if not self.is_dev_mode:
            raise PopConfigurationException(self.pop_id, 'load_model not supported in production mode')
        warnings.warn("load_model for development use only", DeprecationWarning)
        if override:
            await self.unload_model(model)
        if self.sandbox_id is None:
            post_path = f'models/instances'
        else:
            post_path = f'models/instances?sandboxId={self.sandbox_id}'
        response = await self._worker_request_with_retry('POST', url_path_and_query=post_path,
                                                         content_type='application/json', data=json.dumps(model))
        return await response.json()

    async def unload_model(self, model_id: str):
        if not self.is_dev_mode:
            raise PopConfigurationException(self.pop_id, 'unload_model not supported in production mode')
        warnings.warn("purge_model for development use only", DeprecationWarning)
        if self.sandbox_id is None:
            delete_path = 'models/instances/{model["id"]}'
        else:
            delete_path = f'models/instances/{model_id}?sandboxId={self.sandbox_id}'

        await self._worker_request_with_retry('DELETE', url_path_and_query=delete_path)

    '''
    '''

    async def dev_mode_pipeline_base_url(self):
        if self.is_dev_mode:
            return f'{await self.dev_mode_base_url()}/pipelines/{self.worker_config["pipeline_id"]}'
        else:
            pass

    async def upload(self, location: str, params: dict | None = None,
                     on_ready: Callable[[WorkerJob], None] | None = None) -> WorkerJob | SyncWorkerJob:
        job = _UploadJob(location=location, params=params, session=self, on_ready=on_ready,
                         callback=self.metrics_collector)
        await  self._task_start(job.execute())
        return job

    async def upload_stream(self, stream: BinaryIO, mime_type: str, params: dict | None = None,
                            on_ready: Callable[[WorkerJob], None] | None = None) -> WorkerJob | SyncWorkerJob:
        job = _UploadStreamJob(stream, mime_type, params=params, session=self, on_ready=on_ready,
                               callback=self.metrics_collector)
        await self._task_start(job.execute())
        return job

    async def load_from(self, location: str, params: dict | None = None,
                        on_ready: Callable[[WorkerJob], None] | None = None) -> WorkerJob | SyncWorkerJob:
        job = _LoadFromJob(location=location, params=params, session=self, on_ready=on_ready,
                           callback=self.metrics_collector)
        await self._task_start(job.execute())
        return job

    async def dev_mode_base_url(self) -> str:
        if self.worker_config is None:
            await self._reconnect()
        return urljoin(self.eyepop_url, self.worker_config['base_url']).rstrip("/")

    #
    # Implements: _WorkerClientSession
    #

    async def pipeline_get(self, url_path_and_query: str, accept: str | None = None,
                           timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        return await self._pipeline_request_with_retry('GET', url_path_and_query, accept=accept, timeout=timeout)

    async def pipeline_post(self, url_path_and_query: str, accept: str | None = None, open_data: Callable = None,
                            content_type: str | None = None,
                            timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        return await self._pipeline_request_with_retry('POST', url_path_and_query, accept=accept, open_data=open_data,
                                                       content_type=content_type, timeout=timeout)

    async def pipeline_patch(self, url_path_and_query: str, accept: str | None = None, data: Any = None,
                             content_type: str | None = None,
                             timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        def noop():
            pass

        def open_data():
            if isinstance(data, str):
                return StringIO(data)
            else:
                return data

        return await self._pipeline_request_with_retry('PATCH', url_path_and_query, accept=accept, open_data=open_data,
                                                       content_type=content_type, timeout=timeout)

    async def pipeline_delete(self, url_path_and_query: str,
                              timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        return await self._pipeline_request_with_retry('DELETE', url_path_and_query, timeout=timeout)

    async def _pipeline_request_with_retry(self, method: str, url_path_and_query: str, accept: str | None = None,
                                           open_data: Callable = None, content_type: str | None = None,
                                           timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        if self.last_fetch_config_success_time is not None and self.last_fetch_config_success_time < time.time() - FORCE_REFRESH_CONFIG_SECS:
            self.worker_config is None

        retried_re_auth = False
        retried_re_config = False
        start_time = time.time()
        while time.time() - start_time < MAX_RETRY_TIME_SECS:
            if self.worker_config is None:
                await self._reconnect()

            entry = self.load_balancer.next_entry(MAX_RETRY_TIME_SECS + 1)
            if entry is None:
                if not retried_re_config:
                    # pipeline might have just shut down
                    log_requests.debug('no healthy endpoints, about to retry with fresh config')
                    self.worker_config = None
                    retried_re_config = True
                    continue
                else:
                    raise PopNotReachableException(self.pop_id, self.load_balancer.get_debug_status())

            url = f'{entry.base_url}/pipelines/{entry.pipeline_id}/{url_path_and_query}'

            headers = {'Authorization': await self._authorization_header()}
            if accept is not None:
                headers['Accept'] = accept
            if content_type is not None:
                headers['Content-Type'] = content_type
            try:
                log_requests.debug('before %s %s', method, url)
                if open_data is not None:
                    with open_data() as data:
                        if isinstance(data, StringIO):
                            response = await self.client_session.request(method, url, headers=headers,
                                                                         data=data.getvalue(), timeout=timeout)
                        else:
                            response = await self.client_session.request(method, url, headers=headers, data=data,
                                                                         timeout=timeout)
                else:
                    response = await self.client_session.request(method, url, headers=headers, timeout=timeout)

                entry.mark_success()

                log_requests.debug('after %s %s status=%d headers=%s', method, url, response.status, response.headers)

                return response
            except aiohttp.ClientResponseError as e:
                if not retried_re_auth and e.status == 401:
                    # auth token might have just expired
                    log_requests.debug('after %s %s: 401, about to retry with fresh access token', method, url)
                    self.token = None
                    self.expire_token_time = None
                    retried_re_auth = True
                elif e.status == 404:
                    entry.mark_error()
                    log_requests.debug('after %s %s: 404, about to retry fail-over', method, url)
                else:
                    log_requests.exception('unexpected error', e)
                    raise e
            except aiohttp.ClientConnectionError:
                entry.mark_error()
                log_requests.debug('after %s %s: 404, about to retry fail-over', method, url)
            except Exception as e:
                log_requests.exception('unexpected error', e)
                raise e

        raise PopNotReachableException(self.pop_id, f"no success after {MAX_RETRY_TIME_SECS} secs")

    async def _worker_request_with_retry(self, method: str, url_path_and_query: str, accept: str | None = None,
                                         data: Any = None, content_type: str | None = None,
                                         timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        url = f'{await self.dev_mode_base_url()}/{url_path_and_query}'
        return await self.request_with_retry(method, url, accept, data, content_type, timeout)



import json
import pytest
from aioresponses import aioresponses, CallbackResult

from eyepop import EyePopSdk
from tests.worker.base_endpoint_test import BaseEndpointTest


class TestEndpointPostTransform(BaseEndpointTest):
    test_post_transform = "test_post_transform"
    test_new_post_transform = "test_new_post_transform"

    @aioresponses()
    def test_sync_get_post_transform(self, mock: aioresponses):

        self.setup_base_mock(mock)
        # authentication
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_post_transform(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'postTransform': self.test_post_transform}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_post_transform)

        with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                      pop_id=self.test_eyepop_pop_id) as endpoint:
            self.assertBaseMock(mock)
            cur_post_transform = endpoint.get_post_transform()
            self.assertEqual(cur_post_transform, self.test_post_transform)

    @aioresponses()
    def test_sync_set_post_transform(self, mock: aioresponses):
        self.setup_base_mock(mock)
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_post_transform(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'postTransform': self.test_post_transform}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_post_transform)
        
        with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                      pop_id=self.test_eyepop_pop_id) as endpoint:
            self.assertBaseMock(mock)

            transform = self.test_post_transform

            def set_post_transform(url, **kwargs) -> CallbackResult:
                nonlocal transform
                if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                    return CallbackResult(status=401, reason='test auth token expired')
                else:
                    transform = json.loads(kwargs['data'])['transform']
                    return CallbackResult(status=204)

            mock.patch(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}/postTransform',
                       callback=set_post_transform)
            
            endpoint.set_post_transform(self.test_new_post_transform)
            self.assertEqual(endpoint.get_post_transform(), self.test_new_post_transform)
            self.assertEqual(transform, self.test_new_post_transform)


    @aioresponses()
    @pytest.mark.asyncio
    async def test_async_get_post_transform(self, mock: aioresponses):

        self.setup_base_mock(mock)
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_post_transform(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'postTransform': self.test_post_transform}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_post_transform)
        
        async with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                            pop_id=self.test_eyepop_pop_id, is_async=True) as endpoint:
            self.assertBaseMock(mock)
            cur_post_transform = await endpoint.get_post_transform()
            self.assertEqual(cur_post_transform, self.test_post_transform)

    
    @aioresponses()
    @pytest.mark.asyncio
    async def test_async_set_post_transform(self, mock: aioresponses):

        self.setup_base_mock(mock)
        mock.post(f'{self.test_eyepop_url}/authentication/token', status=200, body=json.dumps(
            {'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token}))
        # automatic call to get pop comp and store
        def get_post_transform(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'postTransform': self.test_post_transform}))
        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}',
                    callback=get_post_transform)

        transform = self.test_post_transform

        async with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                            pop_id=self.test_eyepop_pop_id, is_async=True) as endpoint:
            self.assertBaseMock(mock)

            def set_post_transform(url, **kwargs) -> CallbackResult:
                nonlocal transform
                if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                    return CallbackResult(status=401, reason='test auth token expired')
                elif kwargs['headers']['Content-Type'] != 'application/json':
                    return CallbackResult(status=400, reason='unsupported content type')
                else:
                    transform = json.loads(kwargs['data'])['transform']
                    return CallbackResult(status=204)

            mock.patch(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}/postTransform',
                       callback=set_post_transform)
            
            await endpoint.set_post_transform(self.test_new_post_transform)
            self.assertEqual(await endpoint.get_post_transform(), self.test_new_post_transform)
            self.assertEqual(transform, self.test_new_post_transform)


import asyncio
from contextlib import suppress


class Periodic:
    def __init__(self, func, time: float):
        self.func = func
        self.time = time
        self.is_started = False
        self._task = None

    async def start(self):
        if not self.is_started:
            self.is_started = True
            # Start task to call func periodically:
            self._task = asyncio.ensure_future(self._run())

    async def stop(self):
        if self.is_started:
            self.is_started = False
            # Stop task and await it stopped:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    async def _run(self):
        while True:
            await asyncio.sleep(self.time)
            await self.func()


__version__ = '0.22.0'

import eyepop.eyepopsdk
import eyepop.worker.worker_jobs
import eyepop.visualize

EyePopSdk = eyepopsdk.EyePopSdk
Job = eyepop.worker.worker_jobs.WorkerJob
Plot = visualize.EyePopPlot



from asyncio import StreamReader
from typing import Callable, Any, BinaryIO, List, Optional
from urllib.parse import urljoin

import aiohttp
from pydantic.tools import parse_obj_as

from eyepop.client_session import ClientSession
from eyepop.data.data_jobs import DataJob, _UploadStreamJob, _ImportFromJob
from eyepop.data.data_syncify import SyncDataJob
from eyepop.data.data_types import DatasetResponse, DatasetCreate, DatasetUpdate, AssetResponse, Prediction, \
    AssetImport, AutoAnnotate, UserReview, TranscodeMode, ModelResponse, ModelCreate, ModelUpdate
from eyepop.endpoint import Endpoint, log_requests

APPLICATION_JSON = "application/json"


class DataClientSession(ClientSession):
    def __init__(self, delegee: ClientSession, base_url: str):
        self.delegee = delegee
        self.base_url = base_url

    async def request_with_retry(self, method: str, url: str, accept: str | None = None, data: Any = None,
                                 content_type: str | None = None,
                                 timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        url = urljoin(self.base_url, url)
        return await self.delegee.request_with_retry(method, url, accept, data, content_type, timeout)


class DataEndpoint(Endpoint):
    """
    Endpoint to the EyePop.ai Data API.
    """

    def __init__(self, secret_key: str, eyepop_url: str, account_id: str, job_queue_length: int,
                 request_tracer_max_buffer: int):
        super().__init__(secret_key, eyepop_url, job_queue_length, request_tracer_max_buffer)
        self.account_uuid = account_id
        self.data_config = None
        self.add_retry_handler(404, self._retry_404)

    async def _retry_404(self, status_code: int, failed_attempts: int) -> bool:
        if failed_attempts > 1:
            return False
        else:
            log_requests.debug('after 404, about to retry with fresh config')
            self.data_config = None
            return True

    async def _disconnect(self):
        pass

    async def _reconnect(self):
        if self.data_config is not None:
            return
        config_url = f'{self.eyepop_url}/data/config?account_uuid={self.account_uuid}'
        async with await self.request_with_retry("GET", config_url) as resp:
            self.data_config = await resp.json()

    async def data_base_url(self) -> str:
        if self.data_config is None:
            await self._reconnect()
        return urljoin(self.eyepop_url, self.data_config['base_url']).rstrip("/")

    """ Model methods """

    async def list_datasets(self, include_hero_asset: bool = False) -> List[DatasetResponse]:
        get_url = f'{await self.data_base_url()}/datasets?account_uuid={self.account_uuid}&include_hero_asset={include_hero_asset}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(List[DatasetResponse], await resp.json())

    async def create_dataset(self, dataset: DatasetCreate) -> DatasetResponse:
        post_url = f'{await self.data_base_url()}/datasets?account_uuid={self.account_uuid}'
        async with await self.request_with_retry("POST", post_url, content_type=APPLICATION_JSON,
                                                 data=dataset.json()) as resp:
            return parse_obj_as(DatasetResponse, await resp.json())

    async def get_dataset(self, dataset_uuid: str, include_hero_asset: bool = False) -> DatasetResponse:
        get_url = f'{await self.data_base_url()}/datasets?dataset_uuid={dataset_uuid}&include_hero_asset={include_hero_asset}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(DatasetResponse, await resp.json())

    async def update_dataset(self, dataset_uuid: str, dataset: DatasetUpdate) -> DatasetResponse:
        patch_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}'
        async with await self.request_with_retry("PATCH", patch_url, content_type=APPLICATION_JSON,
                                                 data=dataset.json()) as resp:
            return parse_obj_as(DatasetResponse, await resp.json())

    async def delete_dataset(self, dataset_uuid: str) -> None:
        delete_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}'
        async with await self.request_with_retry("DELETE", delete_url) as resp:
            return

    async def freeze_dataset_version(self, dataset_uuid: str, dataset_version: Optional[int] = None) -> DatasetResponse:
        if dataset_version:
            post_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}?dataset_version={dataset_version}'
        else:
            post_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}'
        async with await self.request_with_retry("POST", post_url) as resp:
            return parse_obj_as(DatasetResponse, await resp.json())

    async def delete_dataset_version(self, dataset_uuid: str, dataset_version: int) -> DatasetResponse:
        post_url = f'{await self.data_base_url()}/datasets/{dataset_uuid}/delete?dataset_version={dataset_version}'
        async with await self.request_with_retry("POST", post_url) as resp:
            return parse_obj_as(DatasetResponse, await resp.json())

    """" Asset methods """

    async def upload_asset_job(self, stream: BinaryIO, mime_type: str, dataset_uuid: str,
                               dataset_version: Optional[int] = None, external_id: Optional[str] = None,
                               on_ready: Callable[[DataJob], None] | None = None) -> DataJob | SyncDataJob:
        session = DataClientSession(self, await self.data_base_url())
        job = _UploadStreamJob(stream=stream, mime_type=mime_type, dataset_uuid=dataset_uuid,
                               dataset_version=dataset_version, external_id=external_id, session=session,
                               on_ready=on_ready, callback=self.metrics_collector)
        await self._task_start(job.execute())
        return job

    async def import_asset_job(self, asset_import: AssetImport, dataset_uuid: str, dataset_version: Optional[int] = None,
                               external_id: Optional[str] = None,
                               on_ready: Callable[[DataJob], None] | None = None) -> DataJob | SyncDataJob:
        session = DataClientSession(self, await self.data_base_url())
        job = _ImportFromJob(asset_import=asset_import, dataset_uuid=dataset_uuid, dataset_version=dataset_version,
                             external_id=external_id,session=session,
                             on_ready=on_ready, callback=self.metrics_collector)
        await self._task_start(job.execute())
        return job

    async def list_assets(self, dataset_uuid: str, dataset_version: Optional[int] = None,
                          include_annotations: bool = False) -> List[AssetResponse]:
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        get_url = f'{await self.data_base_url()}/assets?dataset_uuid={dataset_uuid}&include_annotations={"true" if include_annotations else "false"}{version_query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(List[AssetResponse], await resp.json())

    async def get_asset(self, asset_uuid: str, dataset_uuid: Optional[str] = None,
                        dataset_version: Optional[int] = None, include_annotations: bool = False) -> AssetResponse:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        get_url = f'{await self.data_base_url()}/assets/{asset_uuid}?include_annotations={"true" if include_annotations else "false"}{dataset_query}{version_query}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(AssetResponse, await resp.json())

    async def delete_asset(self, asset_uuid: str, dataset_uuid: Optional[str] = None,
                           dataset_version: Optional[int] = None) -> None:
        dataset_query = f'dataset_uuid={dataset_uuid}&' if dataset_uuid is not None else ''
        version_query = f'dataset_version={dataset_version}' if dataset_version is not None else ''
        delete_url = f'{await self.data_base_url()}/assets/{asset_uuid}?{dataset_query}{version_query}'
        async with await self.request_with_retry("DELETE", delete_url) as resp:
            return

    async def resurrect_asset(self, asset_uuid: str, dataset_uuid: str, from_dataset_version: int,
                              into_dataset_version: Optional[int] = None) -> None:
        into_version_query = f'&into_dataset_version={into_dataset_version}' if into_dataset_version is not None else ''
        post_url = f'{await self.data_base_url()}/assets/{asset_uuid}/resurrect?dataset_uuid={dataset_uuid}&from_dataset_version={from_dataset_version}{into_version_query}'
        async with await self.request_with_retry("POST", post_url) as resp:
            return

    async def update_asset_ground_truth(self, asset_uuid: str, dataset_uuid: Optional[str] = None,
                                        dataset_version: Optional[int] = None,
                                        ground_truth: Optional[Prediction] = None) -> None:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        patch_url = f'{await self.data_base_url()}/assets/{asset_uuid}/ground_truth?{dataset_query}{version_query}'
        async with await self.request_with_retry("PATCH", patch_url,
                                                 content_type=APPLICATION_JSON if ground_truth else None,
                                                 data=ground_truth.json() if ground_truth else None) as resp:
            return

    async def update_asset_auto_annotation_status(self, asset_uuid: str, auto_annotate: AutoAnnotate,
                                                  user_review: UserReview, dataset_uuid: Optional[str] = None,
                                                  dataset_version: Optional[int] = None) -> None:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        patch_url = f'{await self.data_base_url()}/assets/{asset_uuid}/auto_annotations/{auto_annotate}/user_review/{user_review}?{dataset_query}{version_query}'
        async with await self.request_with_retry("PATCH", patch_url) as resp:
            return

    async def download_asset(self, asset_uuid: str, dataset_uuid: Optional[str] = None,
                             dataset_version: Optional[int] = None,
                             transcode_mode: TranscodeMode = TranscodeMode.original) -> StreamReader:
        dataset_query = f'&dataset_uuid={dataset_uuid}' if dataset_uuid is not None else ''
        version_query = f'&dataset_version={dataset_version}' if dataset_version is not None else ''
        get_url = f'{await self.data_base_url()}/assets/{asset_uuid}/download?transcode_mode={transcode_mode}{dataset_query}{version_query}'
        resp = await self.request_with_retry("GET", get_url)
        return resp.content

    """ Model methods """

    async def list_models(self) -> List[ModelResponse]:
        get_url = f'{await self.data_base_url()}/models?account_uuid={self.account_uuid}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(List[ModelResponse], await resp.json())

    async def create_model(self, dataset_uuid: str, dataset_version: int, model: ModelCreate) -> ModelResponse:
        post_url = f'{await self.data_base_url()}/models?dataset_uuid={dataset_uuid}&dataset_version={dataset_version}'
        async with await self.request_with_retry("POST", post_url, content_type=APPLICATION_JSON,
                                                 data=model.json()) as resp:
            return parse_obj_as(ModelResponse, await resp.json())

    async def get_model(self, model_uuid: str) -> ModelResponse:
        get_url = f'{await self.data_base_url()}/models/{model_uuid}'
        async with await self.request_with_retry("GET", get_url) as resp:
            return parse_obj_as(ModelResponse, await resp.json())

    async def update_model(self, model_uuid: str, model: ModelUpdate) -> ModelResponse:
        patch_url = f'{await self.data_base_url()}/models/{model_uuid}'
        async with await self.request_with_retry("PATCH", patch_url, content_type=APPLICATION_JSON,
                                                 data=model.json()) as resp:
            return parse_obj_as(ModelResponse, await resp.json())

    async def delete_model(self, model_uuid: str) -> None:
        delete_url = f'{await self.data_base_url()}/models/{model_uuid}'
        async with await self.request_with_retry("DELETE", delete_url) as resp:
            return

    async def publish_model(self, model_uuid: str) -> ModelResponse:
        post_url = f'{await self.data_base_url()}/models/{model_uuid}/publish'
        async with await self.request_with_retry("POST", post_url) as resp:
            return parse_obj_as(ModelResponse, await resp.json())



import asyncio
import logging
import time
import uuid
import flatbuffers
from collections import deque
from types import SimpleNamespace
import importlib.metadata

import aiohttp
from aiohttp import TraceConfig, TraceRequestEndParams, TraceRequestExceptionParams, TraceRequestStartParams, \
    TraceResponseChunkReceivedParams, ClientConnectionError, ServerTimeoutError, ClientResponseError, \
    TraceRequestChunkSentParams

from eyepop.events.Event import EventStart, EventEnd, EventAddMethod, EventAddEventTimeEpochMs, EventAddXRequestId, \
    EventAddResult, EventAddStatus, EventAddHostIndex, EventAddWaitMs, EventAddProcessMs, EventAddPathIndex, \
    EventAddBodyBytesSent, EventAddBodyBytesReceived
from eyepop.events.Record import RecordStart, RecordEnd, RecordAddClientType, RecordAddHosts, \
    RecordAddClientVersion, RecordStartEventsVector, RecordAddEvents, RecordStartPathsVector, RecordAddPaths
from eyepop.events.ClientType import ClientType
from eyepop.events.Record import RecordStartHostsVector
from eyepop.events.Result import Result

log = logging.getLogger('eyepop.tracer')

__version__ = importlib.metadata.version('eyepop')

def delete_nth(d: deque, n: int):
    d.rotate(-n)
    d.popleft()
    d.rotate(n)


def is_event_started_ended_before(event: SimpleNamespace, threshold: float) -> (bool, bool):
    if event.start > threshold:
        return False, False
    last: float = None
    if hasattr(event, 'request_end'):
        last = event.request_end
    if hasattr(event, 'last_chunk_received') and (last is None or event.last_chunk_received > last):
        last = event.last_chunk_received
    if hasattr(event, 'exception') and (last is None or event.exception > last):
        last = event.exception
    return True, (last is not None) and (last <= threshold)


method_to_fb_enum = {
    'other': 0,
    'get': 1,
    'head': 2,
    'options': 3,
    'post': 4,
    'put': 5,
    'patch': 6,
    'delete': 7,
}

class RequestTracer():
    def __init__(self, max_events: int):
        self.events = deque(maxlen=max_events)

    def get_trace_config(self) -> TraceConfig:
        trace_config = aiohttp.TraceConfig()
        trace_config.on_request_start.append(self.on_request_start)
        trace_config.on_request_chunk_sent.append(self.on_request_chunk_sent)
        trace_config.on_response_chunk_received.append(self.on_response_chunk_received)
        trace_config.on_request_exception.append(self.on_request_exception)
        trace_config.on_request_end.append(self.on_request_end)
        return trace_config

    async def send_and_reset(self, url: str, authorization_header: str, secs_to_mature: float | None):
        if secs_to_mature is None or secs_to_mature <= 0.0:
            matured_events = self.events
            self.events = deque(maxlen=self.events.maxlen)
        else:
            threshold = time.time() - secs_to_mature
            matured_events = deque()
            immature_events = deque()
            while len(self.events) > 0:
                event = self.events.pop()
                started_before, ended_before = is_event_started_ended_before(event, threshold)
                if ended_before:
                    matured_events.appendleft(event)
                elif started_before:
                    immature_events.appendleft(event)
                else:
                    immature_events.appendleft(event)
                    immature_events.extendleft(self.events)
                    break

            self.events = immature_events

        log.info('send_and_reset: %d (version: %s)', len(matured_events), __version__)
        if len(matured_events) > 0:
            builder = flatbuffers.Builder()
            host_to_index = {}
            hosts = []
            path_to_index = {}
            paths = []
            client_events = []
            for event in matured_events:
                if not event.host:
                    host_index = -1
                elif event.host in host_to_index.keys():
                    host_index = host_to_index[event.host]
                else:
                    host = builder.CreateString(event.host)
                    hosts.append(host)
                    host_index = len(hosts) - 1
                    host_to_index[event.host] = host_index

                if event.path in path_to_index.keys():
                    path_index = path_to_index[event.path]
                else:
                    path = builder.CreateString(event.path)
                    paths.append(path)
                    path_index = len(paths) - 1
                    path_to_index[event.path] = path_index

                if hasattr(event, 'x_request_id'):
                    x_request_id = builder.CreateString(event.x_request_id)
                else:
                    x_request_id = None
                EventStart(builder)
                EventAddMethod(builder, event.method)
                EventAddEventTimeEpochMs(builder, int(event.realtime * 1000))
                if x_request_id is not None:
                    EventAddXRequestId(builder, x_request_id)
                EventAddResult(builder, event.result)
                if hasattr(event, 'status'):
                    EventAddStatus(builder, event.status)

                EventAddHostIndex(builder, host_index)
                EventAddPathIndex(builder, path_index)

                if hasattr(event, 'exception'):
                    EventAddWaitMs(builder, round((event.exception - event.start) * 1000))
                if hasattr(event, 'request_end'):
                    EventAddWaitMs(builder, round((event.request_end - event.start) * 1000))
                    if hasattr(event, 'last_chunk_received'):
                        EventAddProcessMs(builder, round((event.last_chunk_received - event.request_end) * 1000))

                if hasattr(event, 'bytes_sent'):
                    EventAddBodyBytesSent(builder, event.bytes_sent)

                if hasattr(event, 'bytes_received'):
                    EventAddBodyBytesReceived(builder, event.bytes_received)

                client_event = EventEnd(builder)
                client_events.append(client_event)

            RecordStartHostsVector(builder, len(hosts))
            for host in reversed(hosts):
                builder.PrependSOffsetTRelative(host)
            hosts_vector = builder.EndVector()

            RecordStartPathsVector(builder, len(paths))
            for path in reversed(paths):
                builder.PrependSOffsetTRelative(path)
            paths_vector = builder.EndVector()

            RecordStartEventsVector(builder, len(client_events))
            for client_event in client_events:
                builder.PrependSOffsetTRelative(client_event)
            events_vector = builder.EndVector()

            client_version = builder.CreateString(__version__)

            RecordStart(builder)
            RecordAddClientType(builder, ClientType.python)
            RecordAddClientVersion(builder, client_version)
            RecordAddHosts(builder, hosts_vector)
            RecordAddPaths(builder, paths_vector)
            RecordAddEvents(builder, events_vector)
            record = RecordEnd(builder)
            builder.Finish(record)
            buf = builder.Output()

            async with aiohttp.ClientSession() as session:
                headers = {
                    'authorization': authorization_header,
                    'content-type': 'application/x-flatbuffers;schema=eyepop.events.Record'
                }

                async with session.post(url, data=buf, headers=headers) as resp:
                    pass


    async def on_request_start(self, session, trace_config_ctx: SimpleNamespace, params: TraceRequestStartParams):
        trace_config_ctx.x_request_id = uuid.uuid4().hex
        params.headers.add('X-Request-Id', trace_config_ctx.x_request_id)
        trace_config_ctx.realtime = time.time()
        trace_config_ctx.start = asyncio.get_event_loop().time()
        if params.url.port is not None:
            trace_config_ctx.host = f'{params.url.host}:{params.url.port}'
        else:
            trace_config_ctx.host = params.url.host
        trace_config_ctx.method = method_to_fb_enum[params.method.lower()]
        trace_config_ctx.path = params.url.path
        self.events.appendleft(trace_config_ctx)

    async def on_request_chunk_sent(self, session, trace_config_ctx: SimpleNamespace, params: TraceRequestChunkSentParams):
        bytes_sent = 0
        if hasattr(trace_config_ctx, 'bytes_sent'):
            bytes_sent = trace_config_ctx.bytes_sent
        trace_config_ctx.bytes_sent = bytes_sent + len(params.chunk)

    async def on_response_chunk_received(self, session, trace_config_ctx: SimpleNamespace, params: TraceResponseChunkReceivedParams):
        trace_config_ctx.last_chunk_received = asyncio.get_event_loop().time()
        bytes_received = 0
        if hasattr(trace_config_ctx, 'bytes_received'):
            bytes_received = trace_config_ctx.bytes_received
        trace_config_ctx.bytes_received = bytes_received + len(params.chunk)

    async def on_request_exception(self, session, trace_config_ctx: SimpleNamespace, params: TraceRequestExceptionParams):
        trace_config_ctx.exception = asyncio.get_event_loop().time()
        if isinstance(params.exception, ClientConnectionError):
            trace_config_ctx.result = Result.connection
        elif isinstance(params.exception, ServerTimeoutError):
            trace_config_ctx.result = Result.timeout
        elif isinstance(params.exception, ClientResponseError):
            trace_config_ctx.result = Result.status
        else:
            trace_config_ctx.result = Result.other


    async def on_request_end(self, session, trace_config_ctx: SimpleNamespace, params: TraceRequestEndParams):
        trace_config_ctx.request_end = asyncio.get_event_loop().time()
        trace_config_ctx.status = params.response.status
        trace_config_ctx.result = Result.success




class PopNotStartedException(Exception):
    """
    Thrown when the given Pop was not started and the Endpoint was requested with auto_start=False
    """
    def __init__(self, pop_id: str):
        super().__init__(f'Pop {pop_id} has not been started')


class PopConfigurationException(Exception):
    """
    Thrown when the given Pop configuration was inconsistent 
    """
    def __init__(self, pop_id: str, reason: str):
        super().__init__(f'Pop {pop_id} configuration issue: {reason}')


class PopNotReachableException(Exception):
    """
    Thrown when no healthy endpoint is available
    """
    def __init__(self, pop_id: str, endpoints: list[dict]):
        super().__init__(f'No healthy endpoint for Pop {pop_id} tried: {endpoints}')




import typing
from typing import BinaryIO, Callable, Optional, List

from eyepop.data.data_jobs import DataJob
from eyepop.data.data_types import AssetImport, DatasetResponse, DatasetCreate, DatasetUpdate, AssetResponse, \
    Prediction, AutoAnnotate, UserReview, TranscodeMode, ModelResponse, ModelCreate, ModelUpdate
from eyepop.syncify import run_coro_thread_save, SyncEndpoint, submit_coro_thread_save


class SyncDataJob:
    def __init__(self, job: DataJob, event_loop):
        self.job = job
        self.event_loop = event_loop

    def result(self) -> dict:
        result = run_coro_thread_save(self.event_loop, self.job.result())
        return result

    def cancel(self):
        run_coro_thread_save(self.event_loop, self.job.cancel())


class SyncDataEndpoint(SyncEndpoint):
    def __init__(self, endpoint: "DataEndpoint"):
        super().__init__(endpoint)

    """ Model methods """

    def list_datasets(self, include_hero_asset: bool = False) -> List[DatasetResponse]:
        return run_coro_thread_save(self.event_loop, self.endpoint.list_datasets(include_hero_asset))

    def create_dataset(self, dataset: DatasetCreate) -> DatasetResponse:
        return run_coro_thread_save(self.event_loop, self.endpoint.create_dataset(dataset))

    def get_dataset(self, dataset_uuid: str, include_hero_asset: bool = False) -> DatasetResponse:
        return run_coro_thread_save(self.event_loop, self.get_dataset(dataset_uuid, include_hero_asset))

    def update_dataset(self, dataset_uuid: str, dataset: DatasetUpdate) -> DatasetResponse:
        return run_coro_thread_save(self.event_loop, self.update_dataset(dataset_uuid, dataset))

    def delete_dataset(self, dataset_uuid: str) -> None:
        return run_coro_thread_save(self.event_loop, self.endpoint.delete_dataset(dataset_uuid))

    def freeze_dataset_version(self, dataset_uuid: str, dataset_version: Optional[int] = None) -> DatasetResponse:
        return run_coro_thread_save(self.event_loop,
                                    self.endpoint.freeze_dataset_version(dataset_uuid, dataset_version))

    def delete_dataset_version(self, dataset_uuid: str, dataset_version: int) -> DatasetResponse:
        return run_coro_thread_save(self.event_loop,
                                    self.endpoint.delete_dataset_version(dataset_uuid, dataset_version))

    """" Asset methods """

    def upload_asset_job(self, stream: BinaryIO, mime_type: str, dataset_uuid: str,
                         dataset_version: int | None = None, external_id: str | None = None,
                         on_ready: Callable[[DataJob], None] | None = None) -> DataJob | SyncDataJob:
        if on_ready is not None:
            raise TypeError("'on_ready' callback not supported for sync endpoints. "
                            "Use 'EyePopSdk.dataEndpoint(is_async=True)` to create "
                            "an endpoint with callback support")
        job = run_coro_thread_save(self.event_loop,
                                   self.endpoint.upload_asset_job(stream, mime_type, dataset_uuid, dataset_version,
                                                                   external_id, None))
        return SyncDataJob(job, self.event_loop)

    def import_asset_job(self, asset_import: AssetImport, dataset_uuid: str, dataset_version: Optional[int] = None,
                         external_id: Optional[str] = None,
                         on_ready: Callable[[DataJob], None] | None = None) -> DataJob | SyncDataJob:
        if on_ready is not None:
            raise TypeError("'on_ready' callback not supported for sync endpoints. "
                            "Use 'EyePopSdk.dataEndpoint(is_async=True)` to create "
                            "an endpoint with callback support")
        job = run_coro_thread_save(self.event_loop,
                                   self.endpoint.import_asset_job(asset_import, dataset_uuid, dataset_version,
                                                                   external_id,
                                                                   None))
        return SyncDataJob(job, self.event_loop)

    def list_assets(self, dataset_uuid: str, dataset_version: Optional[int] = None,
                    include_annotations: bool = False) -> List[AssetResponse]:
        return run_coro_thread_save(self.event_loop,
                                    self.endpoint.list_assets(dataset_uuid, dataset_version, include_annotations))

    def get_asset(self, asset_uuid: str, dataset_uuid: Optional[str] = None,
                  dataset_version: Optional[int] = None, include_annotations: bool = False) -> AssetResponse:
        return run_coro_thread_save(self.event_loop, self.endpoint.get_asset(asset_uuid, dataset_uuid, dataset_version,
                                                                             include_annotations))

    def delete_asset(self, asset_uuid: str, dataset_uuid: Optional[str] = None,
                     dataset_version: Optional[int] = None) -> None:
        return run_coro_thread_save(self.event_loop,
                                    self.endpoint.delete_asset(asset_uuid, dataset_uuid, dataset_version))

    def resurrect_asset(self, asset_uuid: str, dataset_uuid: str, from_dataset_version: int,
                        into_dataset_version: Optional[int] = None) -> None:
        return run_coro_thread_save(self.event_loop,
                                    self.endpoint.resurrect_asset(asset_uuid, dataset_uuid, from_dataset_version,
                                                                   into_dataset_version))

    def update_asset_ground_truth(self, asset_uuid: str, dataset_uuid: Optional[str] = None,
                                  dataset_version: Optional[int] = None,
                                  ground_truth: Optional[Prediction] = None) -> None:
        return run_coro_thread_save(self.event_loop,
                                    self.endpoint.update_asset_ground_truth(asset_uuid, dataset_uuid,
                                                                            dataset_version, ground_truth))

    def update_asset_auto_annotation_status(self, asset_uuid: str, auto_annotate: AutoAnnotate,
                                            user_review: UserReview, dataset_uuid: Optional[str] = None,
                                            dataset_version: Optional[int] = None) -> None:
        return run_coro_thread_save(self.event_loop,
                                    self.endpoint.update_asset_auto_annotation_status(asset_uuid, auto_annotate,
                                                                                       user_review, dataset_uuid,
                                                                                       dataset_version))

    def download_asset(self, asset_uuid: str, dataset_uuid: Optional[str] = None,
                       dataset_version: Optional[int] = None,
                       transcode_mode: TranscodeMode = TranscodeMode.original) -> typing.BinaryIO:
        async_stream_reader = run_coro_thread_save(
            self.event_loop,
            self.endpoint.download_asset(asset_uuid, dataset_uuid, dataset_version, transcode_mode)
        )
        sync_io = self._async_reader_to_sync_binary_io(async_stream_reader)
        return sync_io

    """ Model methods """

    def list_models(self) -> List[ModelResponse]:
        return run_coro_thread_save(self.event_loop, self.endpoint.list_models())

    def create_model(self, dataset_uuid: str, dataset_version: int, model: ModelCreate) -> ModelResponse:
        return run_coro_thread_save(self.event_loop, self.endpoint.create_model(dataset_uuid, dataset_version, model))

    def get_model(self, model_uuid: str) -> ModelResponse:
        return run_coro_thread_save(self.event_loop, self.endpoint.get_model(model_uuid))

    def update_model(self, model_uuid: str, model: ModelUpdate) -> ModelResponse:
        return run_coro_thread_save(self.event_loop, self.endpoint.update_model(model_uuid, model))

    def delete_model(self, model_uuid: str) -> None:
        return run_coro_thread_save(self.event_loop, self.endpoint.delete_model(model_uuid))

    def publish_model(self, model_uuid: str) -> ModelResponse:
        return run_coro_thread_save(self.event_loop, self.endpoint.publish_model(model_uuid))



import asyncio
import logging
import time
from pathlib import Path

from eyepop import EyePopSdk

example_file_1 = '/home/torsten/Downloads/remux_pet_owner.mp4.mp4'

logging.basicConfig(level=logging.INFO)
logging.getLogger('eyepop').setLevel(level=logging.DEBUG)


def upload_video(location: str):
    with EyePopSdk.workerEndpoint() as endpoint:
        job = endpoint.upload(location)
        while result := job.predict():
            print(result)



t1 = time.time()
upload_video(example_file_1)
t2 = time.time()
print("1x video async: ", t2 - t1)





import json
import time
import aiohttp
from aioresponses import aioresponses, CallbackResult

from eyepop import EyePopSdk
from tests.worker.base_endpoint_test import BaseEndpointTest


class TestEndpointLoadBalance(BaseEndpointTest):
    test_source_id = 'test_source_id'
    test_url = 'http://examle-media.test/test.png'

    @aioresponses()
    def test_sync_load_one(self, mock: aioresponses):
        self._prepare_mock(mock, 1)
        with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                      pop_id=self.test_eyepop_pop_id) as endpoint:
            self.assertBaseMock(mock, status='active_prod')
            test_timestamp = time.time() * 1000 * 1000 * 1000

            self.test_access_token = '... a new access token ....'
            self._prepare_mock(mock, 1)

            def loadFrom(url, **kwargs) -> CallbackResult:
                nonlocal test_timestamp
                if kwargs['headers']['Authorization'] == f'Bearer {self.test_access_token}':
                    return CallbackResult(status=200,
                                          body=json.dumps(
                                              {'source_id': self.test_source_id, 'seconds': 0,
                                               'system_timestamp': test_timestamp}))
                else:
                    return CallbackResult(status=401, reason='test auth token expired')

            mock.patch(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}-0/source?mode=queue&processing=sync',
                       callback=loadFrom, repeat=True)

            job = endpoint.load_from(self.test_url)
            result = job.predict()
            self.assertIsNotNone(result)
            self.assertEqual(result['source_id'], self.test_source_id)
            self.assertEqual(result['seconds'], 0)
            self.assertEqual(result['system_timestamp'], test_timestamp)
            self.assertIsNone(job.predict())

            mock.assert_called_with(
                f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}-0/source?mode=queue&processing=sync',
                method='PATCH',
                headers={
                    'Accept': 'application/jsonl',
                    'Content-Type': 'application/json',
                    'Authorization': f'Bearer {self.test_access_token}'
                },
                data=json.dumps({'sourceType': 'URL', 'url': self.test_url}),
                timeout=aiohttp.ClientTimeout(total=None, sock_read=60))

    @aioresponses()
    def test_sync_load_two(self, mock: aioresponses):
        self._prepare_mock(mock, 2)
        with EyePopSdk.workerEndpoint(eyepop_url=self.test_eyepop_url, secret_key=self.test_eyepop_secret_key,
                                      pop_id=self.test_eyepop_pop_id) as endpoint:
            self.assertBaseMock(mock, status='active_prod')
            test_timestamp = time.time() * 1000 * 1000 * 1000

            self.test_access_token = '... a new access token ....'
            self._prepare_mock(mock, 2)

            def loadFrom(url, **kwargs) -> CallbackResult:
                nonlocal test_timestamp
                if kwargs['headers']['Authorization'] == f'Bearer {self.test_access_token}':
                    return CallbackResult(status=200,
                                          body=json.dumps(
                                              {'source_id': self.test_source_id, 'seconds': 0,
                                               'system_timestamp': test_timestamp}))
                else:
                    return CallbackResult(status=401, reason='test auth token expired')

            mock.patch(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}-0/source?mode=queue&processing=sync',
                       callback=loadFrom, repeat=True)

            mock.patch(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}-1/source?mode=queue&processing=sync',
                       callback=loadFrom, repeat=True)

            job = endpoint.load_from(self.test_url)
            result = job.predict()
            self.assertIsNotNone(result)
            self.assertEqual(result['source_id'], self.test_source_id)
            self.assertEqual(result['seconds'], 0)
            self.assertEqual(result['system_timestamp'], test_timestamp)
            self.assertIsNone(job.predict())

            job = endpoint.load_from(self.test_url)
            result = job.predict()
            self.assertIsNotNone(result)
            self.assertEqual(result['source_id'], self.test_source_id)
            self.assertEqual(result['seconds'], 0)
            self.assertEqual(result['system_timestamp'], test_timestamp)
            self.assertIsNone(job.predict())

            mock.assert_called_with(
                f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}-0/source?mode=queue&processing=sync',
                method='PATCH',
                headers={
                    'Accept': 'application/jsonl',
                    'Content-Type': 'application/json',
                    'Authorization': f'Bearer {self.test_access_token}'
                },
                data=json.dumps({'sourceType': 'URL', 'url': self.test_url}),
                timeout=aiohttp.ClientTimeout(total=None, sock_read=60))

            mock.assert_called_with(
                f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}-1/source?mode=queue&processing=sync',
                method='PATCH',
                headers={
                    'Accept': 'application/jsonl',
                    'Content-Type': 'application/json',
                    'Authorization': f'Bearer {self.test_access_token}'
                },
                data=json.dumps({'sourceType': 'URL', 'url': self.test_url}),
                timeout=aiohttp.ClientTimeout(total=None, sock_read=60))

    def _prepare_mock(self, mock, num_endpoints):
        self.setup_base_mock(mock, status='active_prod', num_endpoints=num_endpoints)
        mock.post(f'{self.test_eyepop_url}/authentication/token',
                  status=200,
                  body=json.dumps({
                      'expires_in': 1000 * 1000, 'token_type': 'Bearer', 'access_token': self.test_access_token
                  }),
                  repeat=True)

        # automatic call to get pop comp and store
        def get_pop_comp(url, **kwargs) -> CallbackResult:
            if kwargs['headers']['Authorization'] != f'Bearer {self.test_access_token}':
                return CallbackResult(status=401, reason='test auth token expired')
            else:
                return CallbackResult(status=200, body=json.dumps({'inferPipeline': self.test_pop_comp}))

        mock.get(f'{self.test_worker_url}/pipelines/{self.test_pipeline_id}-0',
                 callback=get_pop_comp)



import json
import logging
import mimetypes
from asyncio import Queue
from typing import Callable, BinaryIO

import aiohttp

from eyepop.worker.worker_client_session import (WorkerClientSession)
from eyepop.jobs import Job, JobStateCallback

log_requests = logging.getLogger('eyepop.requests')


class WorkerJob(Job):
    """
    Abstract Job submitted to an EyePop.ai WorkerEndpoint.
    """

    def __init__(self,
                 session: WorkerClientSession,
                 params: dict | None,
                 on_ready: Callable[["WorkerJob"], None] | None,
                 callback: JobStateCallback | None = None):
        super().__init__(session, on_ready, callback)
        self._params = params


    async def predict(self) -> dict:
        return await self.pop_result()

    async def _do_read_response(self, queue: Queue) -> bool:
        got_results = False
        if self._response is not None:
            async with self._response as response:
                self._callback.first_result(self)
                while line := await response.content.readline():

                    # TODO aiohttp should do do this internally
                    for trace in response._traces:
                        await trace.send_response_chunk_received(
                            response.method, response.url, line
                        )

                    if not got_results:
                        got_results = True
                    prediction = json.loads(line)
                    await self.push_result(prediction)
        return got_results


class _UploadJob(WorkerJob):
    def __init__(self, location: str, params: dict | None,
                 session: WorkerClientSession,
                 on_ready: Callable[[WorkerJob], None] | None = None,
                 callback: JobStateCallback | None = None):
        super().__init__(session, params, on_ready, callback)
        self.location = location
        self.target_url = 'source?mode=queue&processing=sync'
        mime_types = mimetypes.guess_type(location)
        if len(mime_types) > 0:
            self.mime_type = mime_types[0]
        else:
            self.mime_type = 'application/octet-stream'

    async def _do_execute_job(self, queue: Queue, session: WorkerClientSession):
        def open_file():
            return open(self.location, 'rb')

        def open_mp_writer():
            mp_writer = aiohttp.MultipartWriter('form-data')
            params_part = mp_writer.append_json(self._params)
            params_part.set_content_disposition('form-data', name='params', filename='blob')
            file_part = mp_writer.append(open_file(), {'Content-Type': self.mime_type})
            file_part.set_content_disposition('form-data', name='file', filename='blob')

        try:
            got_result = False
            if self._params is None:
                self._response = await session.pipeline_post(self.target_url,
                                                             accept='application/jsonl',
                                                             open_data=open_file,
                                                             content_type=self.mime_type,
                                                             timeout=aiohttp.ClientTimeout(total=None, sock_read=60))
            else:
                self._response = await session.pipeline_post(self.target_url,
                                                             accept='application/jsonl',
                                                             open_data=open_mp_writer,
                                                             timeout=aiohttp.ClientTimeout(total=None, sock_read=60))

            got_result = await self._do_read_response(queue)
        finally:
            if not got_result:
                pass
                # await queue.put(None)


class _UploadStreamJob(WorkerJob):
    def __init__(self, stream: BinaryIO, mime_type: str, params: dict | None,
                 session: WorkerClientSession,
                 on_ready: Callable[[WorkerJob], None] | None = None,
                 callback: JobStateCallback | None = None):
        super().__init__(session, params, on_ready, callback)
        self.stream = stream
        self.mime_type = mime_type
        self.target_url = 'source?mode=queue&processing=sync'

    async def _do_execute_job(self, queue: Queue, session: WorkerClientSession):
        if self._params is None:
            self._response = await session.pipeline_post(self.target_url,
                                                         accept='application/jsonl',
                                                         content_type=self.mime_type,
                                                         open_data=lambda: self.stream,
                                                         timeout=aiohttp.ClientTimeout(total=None, sock_read=60))
        else:
            with aiohttp.MultipartWriter('form-data') as mp_writer:
                params_part = mp_writer.append_json(self._params)
                params_part.set_content_disposition('form-data', name='params')
                file_part = mp_writer.append(self.stream, {'Content-Type': self.mime_type})
                file_part.set_content_disposition('form-data', name='file')
                self._response = await session.pipeline_post(self.target_url,
                                                             accept='application/jsonl',
                                                             open_data=lambda: mp_writer,
                                                             timeout=aiohttp.ClientTimeout(total=None, sock_read=60))

        got_result = await self._do_read_response(queue)
        if not got_result:
            await queue.put(None)


class _LoadFromJob(WorkerJob):
    def __init__(self, location: str, params: dict | None,
                 session: WorkerClientSession,
                 on_ready: Callable[[WorkerJob], None] | None = None,
                 callback: JobStateCallback | None = None):
        super().__init__(session, params, on_ready, callback)
        self.location = location
        self.target_url = 'source?mode=queue&processing=sync'
        self.body = {
            "sourceType": "URL",
            "url": self.location,
        }
        if self._params is not None:
            self.body['params'] = self._params

        self.timeouts = aiohttp.ClientTimeout(total=None, sock_read=60)

    async def _do_execute_job(self, queue: Queue, session: WorkerClientSession):
        try:
            got_result = False
            self._response = await session.pipeline_patch(self.target_url,
                                                          accept='application/jsonl',
                                                          data=json.dumps(self.body),
                                                          content_type='application/json',
                                                          timeout=self.timeouts)
            got_result = await self._do_read_response(queue)
        finally:
            if not got_result:
                await queue.put(None)




import matplotlib.patches as patches
import matplotlib.patheffects as path_effects
import matplotlib.pyplot as plt
from matplotlib.axes import Axes


class EyePopPlot:
    def __init__(self, axes: Axes):
        self.axes = axes

    def prediction(self, prediction: dict):
        if prediction is None:
            return
        objects = prediction.get('objects', None)
        if objects is not None:
            for obj in objects:
                self.object(obj)

    def object(self, obj: dict):
        label = self._label(obj)
        min_dim = min(obj['height'], obj['width'])

        corner_size = max(15, min_dim / 5.33333)

        primary_color = (47 / 255, 167 / 255, 215 / 255, 1)
        secondary_color = (148 / 255, 224 / 255, 255 / 255, 1)
        text_color = (255 / 255, 255 / 255, 255 / 255, 1)

        opacity_color = (47 / 255, 167 / 255, 215 / 255, .2)

        x = obj['x']
        y = obj['y']
        w = obj['width']
        h = obj['height']

        # Add Rectangle
        rect = patches.Rectangle((obj['x'], obj['y']), obj['width'], obj['height'], linewidth=1,
                                 edgecolor=opacity_color, facecolor=opacity_color)
        self.axes.add_patch(rect)

        # top left corner
        points = [(x, y + corner_size), (x, y), (x + corner_size, y)]
        rect = patches.Polygon(points, linewidth=1, edgecolor=primary_color, facecolor='none', closed=False)
        self.axes.add_patch(rect)

        # bottom left corner
        points = [(x, y + h - corner_size), (x, y + h), (x + corner_size, y + h)]
        rect = patches.Polygon(points, linewidth=1, edgecolor=primary_color, facecolor='none', closed=False)
        self.axes.add_patch(rect)

        # top right corner
        points = [(x + w - corner_size, y), (x + w, y), (x + w, y + corner_size)]
        rect = patches.Polygon(points, linewidth=1, edgecolor=primary_color, facecolor='none', closed=False)
        self.axes.add_patch(rect)

        # bottom right corner
        points = [(x + w, y + h - corner_size), (x + w, y + h), (x + w - corner_size, y + h)]
        rect = patches.Polygon(points, linewidth=1, edgecolor=primary_color, facecolor='none', closed=False)
        self.axes.add_patch(rect)

        padding = max(min_dim * .02, 5)
        corner_size = corner_size - padding

        # 2nd top left corner
        points = [(x + padding, y + padding + corner_size), (x + padding, y + padding),
                  (x + padding + corner_size, y + padding)]
        rect = patches.Polygon(points, linewidth=1, edgecolor=secondary_color, facecolor='none', closed=False)
        self.axes.add_patch(rect)

        # 2nd bottom left corner
        points = [(x + padding, y - padding + h - corner_size), (x + padding, y - padding + h),
                  (x + padding + corner_size, y - padding + h)]
        rect = patches.Polygon(points, linewidth=1, edgecolor=secondary_color, facecolor='none', closed=False)
        self.axes.add_patch(rect)

        # 2nd top right corner
        points = [(x - padding + w - corner_size, y + padding), (x - padding + w, y + padding),
                  (x - padding + w, y + padding + corner_size)]
        rect = patches.Polygon(points, linewidth=1, edgecolor=secondary_color, facecolor='none', closed=False)
        self.axes.add_patch(rect)

        # 2nd bottom right corner
        points = [(x - padding + w, y - padding + h - corner_size), (x - padding + w, y - padding + h),
                  (x - padding + w - corner_size, y - padding + h)]
        rect = patches.Polygon(points, linewidth=1, edgecolor=secondary_color, facecolor='none', closed=False)
        self.axes.add_patch(rect)

        text = plt.text(obj['x'] + 10 + padding, obj['y'] + 10 + padding, label, fontsize=10, color=text_color,
                        horizontalalignment='left', verticalalignment='top')

        text.set_path_effects([path_effects.Stroke(linewidth=1, foreground=(1, 1, 1, .7)),
                               path_effects.Stroke(linewidth=1, foreground=(0, 0, 0, .7)), path_effects.Normal()])

    def _label(self, obj: dict) -> str:
        label = obj['classLabel']
        if label == 'person':
            if 'objects' in obj:
                for f in obj['objects']:
                    if 'classLabel' in f and f['classLabel'] == 'face':
                        if 'classes' in f:
                            for c in f['classes']:
                                if 'classLabel' in c:
                                    if c['confidence'] == 1:
                                        label = label + "\n" + c['classLabel']
                                    else:
                                        label = label + "\n" + c['classLabel'] + f" {c['confidence'] * 100:.0f}%" + ""
        return label



import logging
import os.path
import sys
import tempfile

from eyepop import EyePopSdk
from eyepop.data.data_types import DatasetCreate, Prediction, PredictedObject, AssetImport

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logging.getLogger('eyepop.requests').setLevel(level=logging.DEBUG)

example_image_path = sys.argv[1]

with (EyePopSdk.dataEndpoint(job_queue_length=1000) as endpoint):
    dataset = endpoint.create_dataset(DatasetCreate(name="test_dataset"))
    try:
        print(dataset)
        asset_import = AssetImport(
            url="https://nmaahc.si.edu/sites/default/files/styles/max_1300x1300/public/images/header/audience-citizen_0.jpg",
            ground_truth=Prediction(source_width=1.0, source_height=1.0, objects=[
                PredictedObject(classLabel="stuff", confidence=0.99, x=0.4, y=0.4, width=0.2, height=0.2)
            ])
        )
        jobs = []
        for i in range(100):
            jobs.append(endpoint.import_asset_job(asset_import=asset_import, dataset_uuid=dataset.uuid, external_id=f"#{i}"))
        for job in jobs:
            asset = job.result()
            print(asset)

        assets = endpoint.list_assets(dataset_uuid=dataset.uuid)
        print(f"found {len(assets)} assets")

        asset = endpoint.get_asset(assets[0].uuid)
        print(asset)

        tmp_dir = tempfile.mkdtemp()
        tmp_file = os.path.join(tmp_dir, f"asset_{asset.uuid}.jpg")
        with open(tmp_file, "wb") as f:
            asset_stream = endpoint.download_asset(asset.uuid)
            f.write(asset_stream.read())
        print(f"downloaded asset {asset.uuid} to {tmp_file}")
    finally:
        endpoint.delete_dataset(dataset.uuid)






import asyncio
from asyncio import Queue
from enum import Enum
from typing import Callable

from eyepop.client_session import ClientSession
from eyepop.worker.worker_client_session import WorkerClientSession


class JobState(Enum):
    CREATED = 1
    STARTED = 2
    IN_PROGRESS = 3
    FINISHED = 4
    FAILED = 5
    DRAINED = 6

    def __repr__(self):
        return self._name_


class JobStateCallback:
    def created(self, job):
        pass

    def started(self, job):
        pass

    def first_result(self, job):
        pass

    def failed(self, job):
        pass

    def finished(self, job):
        pass

    def drained(self, job):
        pass

    def finalized(self, job):
        pass


class Job:
    """
    Abstract Job submitted to an EyePop.ai Endpoint.
    """

    def __init__(self,
                 session: ClientSession,
                 on_ready: Callable[["Job"], None] | None,
                 callback: JobStateCallback | None = None):
        self.on_ready = on_ready
        self._session = session
        self._response = None
        self._queue = asyncio.Queue(maxsize=128)
        if callback is not None:
            self._callback = callback
        else:
            self._callback = JobStateCallback()
        self._callback.created(self)

    def __del__(self):
        self._callback.finalized(self)

    async def push_result(self, result) -> None:
        await self._queue.put(result)

    async def pop_result(self) -> dict:
        queue = self._queue
        if queue is None:
            return None
        else:
            result = await queue.get()
            if result is None:
                self._queue = None
                self._callback.drained(self)
            elif isinstance(result, Exception):
                self._queue = None
                self._callback.drained(self)
                raise result
            return result

    async def cancel(self):
        queue = self._queue
        if queue is None:
            return None
        self._queue = None
        if self._response is not None:
            self._response.close()
        await queue.put(None)

    async def execute(self):
        queue = self._queue
        session = self._session

        self._callback.started(self)

        try:
            await self._do_execute_job(queue, session)
            await queue.put(None)
        except Exception as e:
            if self._queue is None:
                # we got canceled
                pass
            else:
                self._callback.failed(self)
                await queue.put(e)
        finally:
            if self._response is not None:
                response = self._response.close()
                if response is not None:
                    response.release()
            self._callback.finished(self)
            if self.on_ready is not None:
                await self.on_ready(self)

    async def _do_execute_job(self, queue: Queue, session: ClientSession):
        raise NotImplementedError("can't execute abstract jobs")




import sys
import asyncio
import logging
import time

from eyepop import EyePopSdk
from eyepop import Job
from eyepop.worker.worker_endpoint import WorkerEndpoint

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
# logging.getLogger('eyepop').setLevel(level=logging.DEBUG)
logging.getLogger('eyepop.metrics').setLevel(level=logging.DEBUG)


async def async_load_from_photos(endpoint: WorkerEndpoint, urls: list[str]):
    '''
    Async processing of batch uploads - fast and memory efficient
    '''
    sem = asyncio.Semaphore(0)

    async def on_ready(job: Job):
        nonlocal sem
        try:
            while await job.predict() is not None:
                pass
        except Exception as e:
            logging.exception(e)
        finally:
            sem.release()

    n = 0
    for url in urls:
        await endpoint.load_from(url, on_ready=on_ready)
        n += 1
    for i in range(n):
        await sem.acquire()


async def run_test_async(urls: list[str]):
    t0 = time.time()
    async with EyePopSdk.workerEndpoint(is_async=True, pop_id='transient', job_queue_length=10124) as endpoint:
        await endpoint.set_pop_comp('ep_infer model=eyepop-person:EPPersonB1_Person_TorchScriptCpu_float32')
        t1 = time.time()
        await async_load_from_photos(endpoint, urls)
    t2 = time.time()

    print("%d x photo async took %.3f seconds after %.3f seconds connect time\n\n" % (len(example_image_urls), (t2 - t1), (t1 - t0)))

example_image_url = sys.argv[1]
example_image_urls = [example_image_url] * int(sys.argv[2])

asyncio.run(run_test_async(example_image_urls))



import logging
import typing

from eyepop.syncify import run_coro_thread_save, SyncEndpoint
from eyepop.worker.worker_jobs import WorkerJob

if typing.TYPE_CHECKING:
    from eyepop.worker.worker_endpoint import WorkerEndpoint

log = logging.getLogger(__name__)


class SyncWorkerJob:
    def __init__(self, job: WorkerJob, event_loop):
        self.job = job
        self.event_loop = event_loop

    def predict(self) -> dict:
        prediction = run_coro_thread_save(self.event_loop, self.job.predict())
        return prediction

    def cancel(self):
        run_coro_thread_save(self.event_loop, self.job.cancel())


class SyncWorkerEndpoint(SyncEndpoint):
    def __init__(self, endpoint: "WorkerEndpoint"):
        super().__init__(endpoint)

    def upload(self, file_path: str, params: dict | None = None,
               on_ready: typing.Callable[[WorkerJob], None] | None = None) -> SyncWorkerJob:
        if on_ready is not None:
            raise TypeError(
                "'on_ready' callback not supported for sync endpoints. "
                "Use 'EyePopSdk.workerEndpoint(is_async=True)` to create an async endpoint with callback support")
        job = run_coro_thread_save(self.event_loop, self.endpoint.upload(file_path, params, None))
        return SyncWorkerJob(job, self.event_loop)

    def upload_stream(self, stream: typing.BinaryIO, mime_type: str, params: dict | None = None,
                      on_ready: typing.Callable[[WorkerJob], None] | None = None) -> SyncWorkerJob:
        if on_ready is not None:
            raise TypeError(
                "'on_ready' callback not supported for sync endpoints. "
                "Use 'EyePopSdk.workerEndpoint(is_async=True)` to create an async endpoint with callback support")
        job = run_coro_thread_save(self.event_loop, self.endpoint.upload_stream(stream, mime_type, params, None))
        return SyncWorkerJob(job, self.event_loop)

    def load_from(self, location: str, params: dict | None = None,
                  on_ready: typing.Callable[[WorkerJob], None] | None = None) -> SyncWorkerJob:
        if on_ready is not None:
            raise TypeError(
                "'on_ready' callback not supported for sync endpoints. "
                "Use 'EyePopSdk.workerEndpoint(is_async=True)` to create an async endpoint with callback support")
        job = run_coro_thread_save(self.event_loop, self.endpoint.load_from(location, params, None))
        return SyncWorkerJob(job, self.event_loop)

    def get_pop_comp(self) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.get_pop_comp())

    def set_pop_comp(self, popComp: str) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.set_pop_comp(popComp))

    def get_post_transform(self) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.get_post_transform())

    def set_post_transform(self, transform: str) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.set_post_transform(transform))

    '''
    Start Block
    Below methods are not meant for the end user to use directly. They are used by the SDK internally.
    '''

    def list_models(self) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.list_models())

    def get_manifest(self) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.get_manifest())

    def set_manifest(self, manifest: dict) -> None:
        return run_coro_thread_save(self.event_loop, self.endpoint.set_manifest(manifest))

    def load_model(self, model: dict, override: bool = False) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.load_model(model, override))

    def unload_model(self, model_id: str) -> None:
        return run_coro_thread_save(self.event_loop, self.endpoint.unload_model(model_id))

    '''
    End Block
    '''


